/* Simple U-Boot driver for the BCM2708 mini UARTs */
#include <common.h>
#include <watchdog.h>

#include "serial_bcm2708.h"

#define IO_WRITE(addr, val) (*(volatile unsigned int *)(addr) = (val))
#define IO_READ(addr) (*(volatile unsigned int *)(addr))

#define  CONSOLE_PORT CONFIG_CONS_INDEX

static void bcm2708_putc (int portnum, char c);
static int bcm2708_getc (int portnum);
static int bcm2708_tstc (int portnum);

int serial_init (void)
{
	uint32_t baud_rate_reg = ( (VC_SYSTEM_CLOCK_FREQ / (8 * CONFIG_BAUDRATE)) - 1 );
   int      chars_rxed = 0;
   int      printables = 0;

	/*
	 ** First, Enable the GPIO alt function
	 ** IO_WRITE ((GPIO_BASE + 0xC), 0x00000480)
	 ** THIS IS DONE BY VC platform_init function
	 ** code: uint32_t gpio = IO_READ(0x7e20000C);
	 ** code: gpio = (( gpio & 0xFFFFF03F ) | ( 0x00000480 ) );
	 */

	/* Enable the mini UART */
	IO_WRITE ((UART1_BASE + AUX_MU_EN_REG), 0x00000007);
	
	/* Setup the Baud rate */
	IO_WRITE ((UART1_BASE + AUX_MU_BAUD), baud_rate_reg);
	
	/* Setup the data to be 8 bit */
	IO_WRITE ((UART1_BASE + AUX_MU_LCR_REG), BIT_0 | BIT_1 );

// JACKSON Babble-detect on RX shutoff
// When we're plugged directly into a device, the UART_RX pin
// may receive lots of noise which would appear to be a
// stream of incoming characters.  This confuses UBOOT and
// will often prevent the device from booting at all.
// When we disable the UART RX, this will allow UBOOT to
// continue .

   /* push 16 characters out the port to induce reflections
    * back to our UART RX Pin
    */
   serial_puts ("uboot:0123456789ABCDEF\n");

// TODO: This is a "hammer" which shuts off RX. remove and improve smart detect
//IO_WRITE ((UART1_BASE + AUX_MU_CNTL_REG), 0x00000002);

	while (serial_tstc())
   {
      int      sc;

      sc = serial_getc();

      ++chars_rxed;

      if ( ('\n' == sc) || 
           ('\r' == sc) ||
           ( ' ' == sc) ||
           ((sc >= 'a') && (sc <='z')) ||
           ((sc >= 'A') && (sc <='Z')) ||
           ((sc >= '0') && (sc <='9')) )
      {
         printables++;
      }
   }

   if ( chars_rxed && (printables <= (chars_rxed >> 1)) )
   {
      // Disables receiver functionality on UART as we appear
      // to be getting garbage on the RX Line.
      IO_WRITE ((UART1_BASE + AUX_MU_CNTL_REG), 0x00000002);
      serial_puts ("uboot:UART_RX Babbling " );
      serial_putc ( '0' + chars_rxed );
      serial_putc ( '0' + printables );
      serial_putc ( '\n' );
   }
   else
   {
      serial_puts ("uboot:UART_RX NOT Babbling " );
      serial_putc ( '0' + chars_rxed );
      serial_putc ( '0' + printables );
      serial_putc ( '\n' );
   }
	return 0;
}

void serial_putc (const char c)
{
	if (c == '\n')
		bcm2708_putc (CONSOLE_PORT, '\r');

	bcm2708_putc (CONSOLE_PORT, c);
}

void serial_puts (const char *s)
{
	while (*s) {
		serial_putc (*s++);
	}
}

extern unsigned _stg2_secure_boot;

int serial_getc (void)
{
	if (_stg2_secure_boot & 1) {
		while (1) ;
	}
	return bcm2708_getc (CONSOLE_PORT);
}

int serial_tstc (void)
{
	if (_stg2_secure_boot & 1) {
		return 0;
	}
	return bcm2708_tstc (CONSOLE_PORT);
}

void serial_setbrg (void)
{
}

static void bcm2708_putc (int portnum, char c)
{
   int         timeout = 0;

	/* Wait until there is space in the FIFO */
	while ( !(IO_READ (UART_STATUS_REG) & UART_STATUS_TX_READY))
   {
		WATCHDOG_RESET();
      if ( ++timeout > 1000 )
      {
         return;
      }
   }

	/* Send the character */
	IO_WRITE (UART_DATA, c);
}

static int bcm2708_getc (int portnum)
{
	unsigned int data;

   // JACKSON Babble-detect shutoff
   // if bit 0 is NOT set, RX has been disabled during init, return now
   // as we will never receive another character over RX
   if ( !(0x1 & IO_READ ((UART1_BASE + AUX_MU_CNTL_REG))) )
      return(0);

	/* Wait until there is data in the FIFO */
	while ( !(IO_READ (UART_STATUS_REG) & UART_STATUS_RX_READY))
   {
		WATCHDOG_RESET();
   }

	data = IO_READ (UART_DATA);

	return (int) data;
}

static int bcm2708_tstc (int portnum)
{
   // JACKSON Babble-detect shutoff
   // if bit 0 is NOT set, RX has been disabled during init, return now
   // as we will never receive another character over RX
   if ( !(0x1 & IO_READ ((UART1_BASE + AUX_MU_CNTL_REG))) )
      return(0);

	WATCHDOG_RESET();
	return (IO_READ (UART_STATUS_REG) & UART_STATUS_RX_READY );
}
